/**
 * \file doc_mainpage.h
 *
 * \brief Main page documentation file.
 */
/*
 *
 *  Copyright The Mbed TLS Contributors
 *  SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 */

/**
 * @mainpage Mbed TLS v2.28.8 API Documentation
 *
 * This documentation describes the internal structure of Mbed TLS.  It was
 * automatically generated from specially formatted comment blocks in
 * Mbed TLS's source code using Doxygen.  (See
 * https://www.doxygen.nl for more information on Doxygen)
 */
