#!/bin/sh
set -eu

# Configure and build deps
./configure-mbedtls.sh
./configure-pkcs11-helper.sh
make mbedtls pkcs11-helper

# Configure and build openvpn
./configure-openvpn.sh
make openvpn-nl

# Archive the build artifacts
. "$(dirname $0)/archive.sh"
archive_artifacts
