/*
 * unix2dos.js  - a simple unix2dos implementation in jscript.
 *
 * Copyright (C) 2008-2012 Alon Bar-Lev <alon.barlev@gmail.com>
 *
 * BSD License
 * ============
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     o Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     o Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     o Neither the name of the Alon Bar-Lev nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

var ForReading = 1;

try {
	var fso = new ActiveXObject("Scripting.FileSystemObject");
	for (var i=0;i<WScript.Arguments.length;i++) {
		var file = WScript.Arguments(i);
		var tmp = file+".tmp";
		var fin = fso.OpenTextFile(file, ForReading);
		var fout = fso.CreateTextFile(tmp);
		fout.Write(fin.ReadAll().replace(/\n/g, "\r\n"));
		fin.Close();
		fout.Close();
		fso.DeleteFile(file);
		fso.MoveFile(tmp, file);
	}
	WScript.Quit(0);
}
catch(e) {
	WScript.Echo("ERROR: " + e.description);
	WScript.Quit(1);
}
