#!/bin/sh
set -eu

# Configure and build deps
./configure-mbedtls-win.sh
./configure-pkcs11-helper-win.sh
make mbedtls pkcs11-helper

# Configure and build openvpn
./configure-openvpn-win.sh
make openvpn-nl

# Archive the build artifacts
. "$(dirname $0)/archive.sh"
archive_artifacts "build-openvpn-nl/src/openvpnserv/openvpnserv.exe build-openvpn-nl/src/tapctl/tapctl.exe build-openvpn-nl/doc/openvpn.8.html"
