#!/bin/sh
set -eu

. $(pwd)/nix_build_vars

(cd openvpn && autoreconf -vi)

mkdir -p build-openvpn-openssl
(cd build-openvpn-openssl
	CFLAGS="${CFLAGS:--g -O2 ${HARDENING_FLAGS}}" \
	../openvpn/configure --with-crypto-library=openssl \
        --enable-iproute2 --disable-ofb-cfb
)
